@@ -13,12 +13,12 @@ public class UTF8JsonGenerator
     private final static byte BYTE_u = (byte) 'u';
 
     private final static byte BYTE_0 = (byte) '0';
-    
+
     private final static byte BYTE_LBRACKET = (byte) '[';
     private final static byte BYTE_RBRACKET = (byte) ']';
     private final static byte BYTE_LCURLY = (byte) '{';
     private final static byte BYTE_RCURLY = (byte) '}';
- 
+
     private final static byte BYTE_BACKSLASH = (byte) '\\';
     private final static byte BYTE_COMMA = (byte) ',';
     private final static byte BYTE_COLON = (byte) ':';
@@ -26,8 +26,8 @@ public class UTF8JsonGenerator
 
     // intermediate copies only made up to certain length...
     private final static int MAX_BYTES_TO_BUFFER = 512;
-    
-    final static byte[] HEX_CHARS = CharTypes.copyHexBytes();
+
+    private final static byte[] HEX_CHARS = CharTypes.copyHexBytes();
 
     private final static byte[] NULL_BYTES = { 'n', 'u', 'l', 'l' };
     private final static byte[] TRUE_BYTES = { 't', 'r', 'u', 'e' };
@@ -38,7 +38,7 @@ public class UTF8JsonGenerator
     /* Output buffering
     /**********************************************************
      */
-    
+
     /**
      * Underlying output stream used for writing JSON content.
      */
@@ -54,7 +54,7 @@ public class UTF8JsonGenerator
      * Pointer to the position right beyond the last character to output
      * (end marker; may be past the buffer)
      */
-    protected int _outputTail = 0;
+    protected int _outputTail;
 
     /**
      * End marker of the output buffer; one past the last valid position
@@ -67,18 +67,18 @@ public class UTF8JsonGenerator
      * in the output buffer after escaping
      */
     protected final int _outputMaxContiguous;
-    
+
     /**
      * Intermediate buffer in which characters of a String are copied
      * before being encoded.
      */
     protected char[] _charBuffer;
-    
+
     /**
      * Length of <code>_charBuffer</code>
      */
     protected final int _charBufferLength;
-    
+
     /**
      * 6 character temporary buffer allocated if needed, for constructing
      * escape sequences
@@ -91,18 +91,6 @@ public class UTF8JsonGenerator
      */
     protected boolean _bufferRecyclable;
 
-    /*
-    /**********************************************************
-    /* Quick flags
-    /**********************************************************
-     */
-
-    /**
-     * Flag that is set if quoting is not to be added around
-     * JSON Object property names.
-     */
-    protected boolean _cfgUnqNames;
-
     /*
     /**********************************************************
     /* Life-cycle
@@ -130,7 +118,6 @@ public class UTF8JsonGenerator
         if (isEnabled(Feature.ESCAPE_NON_ASCII)) {
             setHighestNonEscapedChar(127);
         }
-        _cfgUnqNames = !Feature.QUOTE_FIELD_NAMES.enabledIn(features);
     }
     
     public UTF8JsonGenerator(IOContext ctxt, int features, ObjectCodec codec,
@@ -145,10 +132,9 @@ public class UTF8JsonGenerator
         _outputBuffer = outputBuffer;
         _outputEnd = _outputBuffer.length;
         // up to 6 bytes per char (see above), rounded up to 1/8
-        _outputMaxContiguous = _outputEnd >> 3;
+        _outputMaxContiguous = (_outputEnd >> 3);
         _charBuffer = ctxt.allocConcatBuffer();
         _charBufferLength = _charBuffer.length;
-        _cfgUnqNames = !Feature.QUOTE_FIELD_NAMES.enabledIn(features);
     }
 
     /*
@@ -162,6 +148,12 @@ public class UTF8JsonGenerator
         return _outputStream;
     }
 
+    @Override
+    public int getOutputBuffered() {
+        // Assuming tail is always valid, set to 0 on close
+        return _outputTail;
+    }
+
     /*
     /**********************************************************
     /* Overridden methods
@@ -202,15 +194,14 @@ public class UTF8JsonGenerator
             _flushBuffer();
         }
         _outputBuffer[_outputTail++] = BYTE_QUOTE;
-        name.getChars(0, len, _charBuffer, 0);
         // But as one segment, or multiple?
         if (len <= _outputMaxContiguous) {
             if ((_outputTail + len) > _outputEnd) { // caller must ensure enough space
                 _flushBuffer();
             }
-            _writeStringSegment(_charBuffer, 0, len);
+            _writeStringSegment(name, 0, len);
         } else {
-            _writeStringSegments(_charBuffer, 0, len);
+            _writeStringSegments(name, 0, len);
         }
         // and closing quotes; need room for one more char:
         if (_outputTail >= _outputEnd) {
@@ -300,7 +291,7 @@ public class UTF8JsonGenerator
             }
             _outputBuffer[_outputTail++] = BYTE_RBRACKET;
         }
-        _writeContext = _writeContext.getParent();
+        _writeContext = _writeContext.clearAndGetParent();
     }
 
     @Override
@@ -332,7 +323,7 @@ public class UTF8JsonGenerator
             }
             _outputBuffer[_outputTail++] = BYTE_RCURLY;
         }
-        _writeContext = _writeContext.getParent();
+        _writeContext = _writeContext.clearAndGetParent();
     }
 
     /**
@@ -416,45 +407,22 @@ public class UTF8JsonGenerator
     @Override
     public void writeString(String text) throws IOException
     {
-        _verifyValueWrite("write text value");
+        _verifyValueWrite(WRITE_STRING);
         if (text == null) {
             _writeNull();
             return;
         }
-        // First: can we make a local copy of chars that make up text?
+        // First: if we can't guarantee it all fits, quoted, within output, offline
         final int len = text.length();
-        if (len > _charBufferLength) { // nope: off-line handling
+        if (len > _outputMaxContiguous) { // nope: off-line handling
             _writeStringSegments(text, true);
             return;
         }
-        // yes: good.
-        text.getChars(0, len, _charBuffer, 0);
-        // Output: if we can't guarantee it fits in output buffer, off-line as well:
-        if (len > _outputMaxContiguous) {
-            _writeLongString(_charBuffer, 0, len);
-            return;
-        }
         if ((_outputTail + len) >= _outputEnd) {
             _flushBuffer();
         }
         _outputBuffer[_outputTail++] = BYTE_QUOTE;
-        _writeStringSegment(_charBuffer, 0, len); // we checked space already above
-        /* [JACKSON-462] But that method may have had to expand multi-byte Unicode
-         *   chars, so we must check again
-         */
-        if (_outputTail >= _outputEnd) {
-            _flushBuffer();
-        }
-        _outputBuffer[_outputTail++] = BYTE_QUOTE;
-    }
-
-    private void _writeLongString(char[] text, int offset, int len) throws IOException
-    {
-        if (_outputTail >= _outputEnd) {
-            _flushBuffer();
-        }
-        _outputBuffer[_outputTail++] = BYTE_QUOTE;
-        _writeStringSegments(_charBuffer, 0, len);
+        _writeStringSegment(text, 0, len); // we checked space already above
         if (_outputTail >= _outputEnd) {
             _flushBuffer();
         }
@@ -464,7 +432,7 @@ public class UTF8JsonGenerator
     @Override
     public void writeString(char[] text, int offset, int len) throws IOException
     {
-        _verifyValueWrite("write text value");
+        _verifyValueWrite(WRITE_STRING);
         if (_outputTail >= _outputEnd) {
             _flushBuffer();
         }
@@ -488,7 +456,7 @@ public class UTF8JsonGenerator
     @Override
     public final void writeString(SerializableString text) throws IOException
     {
-        _verifyValueWrite("write text value");
+        _verifyValueWrite(WRITE_STRING);
         if (_outputTail >= _outputEnd) {
             _flushBuffer();
         }
@@ -504,11 +472,11 @@ public class UTF8JsonGenerator
         }
         _outputBuffer[_outputTail++] = BYTE_QUOTE;
     }
-
+    
     @Override
     public void writeRawUTF8String(byte[] text, int offset, int length) throws IOException
     {
-        _verifyValueWrite("write text value");
+        _verifyValueWrite(WRITE_STRING);
         if (_outputTail >= _outputEnd) {
             _flushBuffer();
         }
@@ -523,7 +491,7 @@ public class UTF8JsonGenerator
     @Override
     public void writeUTF8String(byte[] text, int offset, int len) throws IOException
     {
-        _verifyValueWrite("write text value");
+        _verifyValueWrite(WRITE_STRING);
         if (_outputTail >= _outputEnd) {
             _flushBuffer();
         }
@@ -590,7 +558,7 @@ public class UTF8JsonGenerator
     // since 2.5
     @Override
     public void writeRawValue(SerializableString text) throws IOException {
-        _verifyValueWrite("write raw value");
+        _verifyValueWrite(WRITE_RAW);
         byte[] raw = text.asUnquotedUTF8();
         if (raw.length > 0) {
             _writeBytes(raw);
@@ -710,7 +678,7 @@ public class UTF8JsonGenerator
             byte[] data, int offset, int len)
         throws IOException, JsonGenerationException
     {
-        _verifyValueWrite("write binary value");
+        _verifyValueWrite(WRITE_BINARY);
         // Starting quotes
         if (_outputTail >= _outputEnd) {
             _flushBuffer();
@@ -729,7 +697,7 @@ public class UTF8JsonGenerator
             InputStream data, int dataLength)
         throws IOException, JsonGenerationException
     {
-        _verifyValueWrite("write binary value");
+        _verifyValueWrite(WRITE_BINARY);
         // Starting quotes
         if (_outputTail >= _outputEnd) {
             _flushBuffer();
@@ -765,10 +733,9 @@ public class UTF8JsonGenerator
      */
 
     @Override
-    public void writeNumber(short s)
-        throws IOException, JsonGenerationException
+    public void writeNumber(short s) throws IOException
     {
-        _verifyValueWrite("write number");
+        _verifyValueWrite(WRITE_NUMBER);
         // up to 5 digits and possible minus sign
         if ((_outputTail + 6) >= _outputEnd) {
             _flushBuffer();
@@ -790,10 +757,9 @@ public class UTF8JsonGenerator
     } 
     
     @Override
-    public void writeNumber(int i)
-        throws IOException, JsonGenerationException
+    public void writeNumber(int i) throws IOException
     {
-        _verifyValueWrite("write number");
+        _verifyValueWrite(WRITE_NUMBER);
         // up to 10 digits and possible minus sign
         if ((_outputTail + 11) >= _outputEnd) {
             _flushBuffer();
@@ -816,10 +782,9 @@ public class UTF8JsonGenerator
     }    
 
     @Override
-    public void writeNumber(long l)
-        throws IOException, JsonGenerationException
+    public void writeNumber(long l) throws IOException
     {
-        _verifyValueWrite("write number");
+        _verifyValueWrite(WRITE_NUMBER);
         if (_cfgNumbersAsStrings) {
             _writeQuotedLong(l);
             return;
@@ -842,14 +807,13 @@ public class UTF8JsonGenerator
     }
 
     @Override
-    public void writeNumber(BigInteger value)
-        throws IOException, JsonGenerationException
+    public void writeNumber(BigInteger value) throws IOException
     {
-        _verifyValueWrite("write number");
+        _verifyValueWrite(WRITE_NUMBER);
         if (value == null) {
             _writeNull();
         } else if (_cfgNumbersAsStrings) {
-            _writeQuotedRaw(value);
+            _writeQuotedRaw(value.toString());
         } else {
             writeRaw(value.toString());
         }
@@ -857,48 +821,46 @@ public class UTF8JsonGenerator
 
     
     @Override
-    public void writeNumber(double d)
-        throws IOException, JsonGenerationException
+    public void writeNumber(double d) throws IOException
     {
         if (_cfgNumbersAsStrings ||
-            // [JACKSON-139]
             (((Double.isNaN(d) || Double.isInfinite(d))
-                && isEnabled(Feature.QUOTE_NON_NUMERIC_NUMBERS)))) {
+                && Feature.QUOTE_NON_NUMERIC_NUMBERS.enabledIn(_features)))) {
             writeString(String.valueOf(d));
             return;
         }
         // What is the max length for doubles? 40 chars?
-        _verifyValueWrite("write number");
+        _verifyValueWrite(WRITE_NUMBER);
         writeRaw(String.valueOf(d));
     }
 
     @Override
-    public void writeNumber(float f)
-        throws IOException, JsonGenerationException
+    public void writeNumber(float f) throws IOException
     {
         if (_cfgNumbersAsStrings ||
             // [JACKSON-139]
             (((Float.isNaN(f) || Float.isInfinite(f))
-                && isEnabled(Feature.QUOTE_NON_NUMERIC_NUMBERS)))) {
+                && Feature.QUOTE_NON_NUMERIC_NUMBERS.enabledIn(_features)))) {
             writeString(String.valueOf(f));
             return;
         }
         // What is the max length for floats?
-        _verifyValueWrite("write number");
+        _verifyValueWrite(WRITE_NUMBER);
         writeRaw(String.valueOf(f));
     }
 
     @Override
-    public void writeNumber(BigDecimal value)
-        throws IOException, JsonGenerationException
+    public void writeNumber(BigDecimal value) throws IOException
     {
         // Don't really know max length for big decimal, no point checking
-        _verifyValueWrite("write number");
+        _verifyValueWrite(WRITE_NUMBER);
         if (value == null) {
             _writeNull();
         } else if (_cfgNumbersAsStrings) {
-            _writeQuotedRaw(value);
-        } else if (isEnabled(Feature.WRITE_BIGDECIMAL_AS_PLAIN)) {
+            String raw = Feature.WRITE_BIGDECIMAL_AS_PLAIN.enabledIn(_features)
+                    ? value.toPlainString() : value.toString();
+            _writeQuotedRaw(raw);
+        } else if (Feature.WRITE_BIGDECIMAL_AS_PLAIN.enabledIn(_features)) {
             writeRaw(value.toPlainString());
         } else {
             writeRaw(value.toString());
@@ -906,10 +868,9 @@ public class UTF8JsonGenerator
     }
 
     @Override
-    public void writeNumber(String encodedValue)
-        throws IOException, JsonGenerationException
+    public void writeNumber(String encodedValue) throws IOException
     {
-        _verifyValueWrite("write number");
+        _verifyValueWrite(WRITE_NUMBER);
         if (_cfgNumbersAsStrings) {
             _writeQuotedRaw(encodedValue);            
         } else {
@@ -917,13 +878,13 @@ public class UTF8JsonGenerator
         }
     }
 
-    private final void _writeQuotedRaw(Object value) throws IOException
+    private final void _writeQuotedRaw(String value) throws IOException
     {
         if (_outputTail >= _outputEnd) {
             _flushBuffer();
         }
         _outputBuffer[_outputTail++] = BYTE_QUOTE;
-        writeRaw(value.toString());
+        writeRaw(value);
         if (_outputTail >= _outputEnd) {
             _flushBuffer();
         }
@@ -931,10 +892,9 @@ public class UTF8JsonGenerator
     }
     
     @Override
-    public void writeBoolean(boolean state)
-        throws IOException, JsonGenerationException
+    public void writeBoolean(boolean state) throws IOException
     {
-        _verifyValueWrite("write boolean value");
+        _verifyValueWrite(WRITE_BOOLEAN);
         if ((_outputTail + 5) >= _outputEnd) {
             _flushBuffer();
         }
@@ -945,10 +905,9 @@ public class UTF8JsonGenerator
     }
 
     @Override
-    public void writeNull()
-        throws IOException, JsonGenerationException
+    public void writeNull() throws IOException
     {
-        _verifyValueWrite("write null value");
+        _verifyValueWrite(WRITE_NULL);
         _writeNull();
     }
 
@@ -959,8 +918,7 @@ public class UTF8JsonGenerator
      */
 
     @Override
-    protected final void _verifyValueWrite(String typeMsg)
-        throws IOException, JsonGenerationException
+    protected final void _verifyValueWrite(String typeMsg) throws IOException
     {
         int status = _writeContext.writeValue();
         if (status == JsonWriteContext.STATUS_EXPECT_NAME) {
@@ -998,8 +956,7 @@ public class UTF8JsonGenerator
         _verifyPrettyValueWrite(typeMsg, status);
     }
 
-    protected final void _verifyPrettyValueWrite(String typeMsg, int status)
-        throws IOException, JsonGenerationException
+    protected final void _verifyPrettyValueWrite(String typeMsg, int status) throws IOException
     {
         // If we have a pretty printer, it knows what to do:
         switch (status) {
@@ -1033,8 +990,7 @@ public class UTF8JsonGenerator
      */
 
     @Override
-    public void flush()
-        throws IOException
+    public void flush() throws IOException
     {
         _flushBuffer();
         if (_outputStream != null) {
@@ -1045,8 +1001,7 @@ public class UTF8JsonGenerator
     }
 
     @Override
-    public void close()
-        throws IOException
+    public void close() throws IOException
     {
         super.close();
 
@@ -1054,7 +1009,7 @@ public class UTF8JsonGenerator
          *   scopes.
          */
         // First: let's see that we still have buffers...
-        if (_outputBuffer != null
+        if ((_outputBuffer != null)
             && isEnabled(Feature.AUTO_CLOSE_JSON_CONTENT)) {
             while (true) {
                 JsonStreamContext ctxt = getOutputContext();
@@ -1068,6 +1023,7 @@ public class UTF8JsonGenerator
             }
         }
         _flushBuffer();
+        _outputTail = 0; // just to ensure we don't think there's anything buffered
 
         /* 25-Nov-2008, tatus: As per [JACKSON-16] we are not to call close()
          *   on the underlying Reader, unless we "own" it, or auto-closing
@@ -1078,7 +1034,7 @@ public class UTF8JsonGenerator
         if (_outputStream != null) {
             if (_ioContext.isResourceManaged() || isEnabled(Feature.AUTO_CLOSE_TARGET)) {
                 _outputStream.close();
-            } else  if (isEnabled(Feature.FLUSH_PASSED_TO_STREAM)) {
+            } else if (isEnabled(Feature.FLUSH_PASSED_TO_STREAM)) {
                 // If we can't close it, we should at least flush
                 _outputStream.flush();
             }
@@ -1161,15 +1117,13 @@ public class UTF8JsonGenerator
 
         int left = text.length();
         int offset = 0;
-        final char[] cbuf = _charBuffer;
 
         while (left > 0) {
             int len = Math.min(_outputMaxContiguous, left);
-            text.getChars(offset, offset+len, cbuf, 0);
             if ((_outputTail + len) > _outputEnd) { // caller must ensure enough space
                 _flushBuffer();
             }
-            _writeStringSegment(cbuf, 0, len);
+            _writeStringSegment(text, offset, len);
             offset += len;
             left -= len;
         }
@@ -1188,8 +1142,7 @@ public class UTF8JsonGenerator
      * the output buffer. If so, we will need to choose smaller output
      * chunks to write at a time.
      */
-    private final void _writeStringSegments(char[] cbuf, int offset, int totalLen)
-        throws IOException, JsonGenerationException
+    private final void _writeStringSegments(char[] cbuf, int offset, int totalLen) throws IOException
     {
         do {
             int len = Math.min(_outputMaxContiguous, totalLen);
@@ -1202,6 +1155,19 @@ public class UTF8JsonGenerator
         } while (totalLen > 0);
     }
 
+    private final void _writeStringSegments(String text, int offset, int totalLen) throws IOException
+    {
+        do {
+            int len = Math.min(_outputMaxContiguous, totalLen);
+            if ((_outputTail + len) > _outputEnd) { // caller must ensure enough space
+                _flushBuffer();
+            }
+            _writeStringSegment(text, offset, len);
+            offset += len;
+            totalLen -= len;
+        } while (totalLen > 0);
+    }
+
     /*
     /**********************************************************
     /* Internal methods, low-level writing, text segments
@@ -1217,7 +1183,7 @@ public class UTF8JsonGenerator
      * potentially enough space for other cases (but not necessarily flushed)
      */
     private final void _writeStringSegment(char[] cbuf, int offset, int len)
-        throws IOException, JsonGenerationException
+        throws IOException
     {
         // note: caller MUST ensure (via flushing) there's room for ASCII only
         
@@ -1252,12 +1218,42 @@ public class UTF8JsonGenerator
         }
     }
 
+    private final void _writeStringSegment(String text, int offset, int len) throws IOException
+    {
+        // note: caller MUST ensure (via flushing) there's room for ASCII only
+        // Fast+tight loop for ASCII-only, no-escaping-needed output
+        len += offset; // becomes end marker, then
+
+        int outputPtr = _outputTail;
+        final byte[] outputBuffer = _outputBuffer;
+        final int[] escCodes = _outputEscapes;
+
+        while (offset < len) {
+            int ch = text.charAt(offset);
+            // note: here we know that (ch > 0x7F) will cover case of escaping non-ASCII too:
+            if (ch > 0x7F || escCodes[ch] != 0) {
+                break;
+            }
+            outputBuffer[outputPtr++] = (byte) ch;
+            ++offset;
+        }
+        _outputTail = outputPtr;
+        if (offset < len) {
+            if (_characterEscapes != null) {
+                _writeCustomStringSegment2(text, offset, len);
+            } else if (_maximumNonEscapedChar == 0) {
+                _writeStringSegment2(text, offset, len);
+            } else {
+                _writeStringSegmentASCII2(text, offset, len);
+            }
+        }
+    }
+
     /**
      * Secondary method called when content contains characters to escape,
      * and/or multi-byte UTF-8 characters.
      */
-    private final void _writeStringSegment2(final char[] cbuf, int offset, final int end)
-        throws IOException, JsonGenerationException
+    private final void _writeStringSegment2(final char[] cbuf, int offset, final int end) throws IOException
     {
         // Ok: caller guarantees buffer can have room; but that may require flushing:
         if ((_outputTail +  6 * (end - offset)) > _outputEnd) {
@@ -1296,6 +1292,44 @@ public class UTF8JsonGenerator
         _outputTail = outputPtr;
     }
 
+    private final void _writeStringSegment2(final String text, int offset, final int end) throws IOException
+    {
+        if ((_outputTail +  6 * (end - offset)) > _outputEnd) {
+            _flushBuffer();
+        }
+
+        int outputPtr = _outputTail;
+
+        final byte[] outputBuffer = _outputBuffer;
+        final int[] escCodes = _outputEscapes;
+        
+        while (offset < end) {
+            int ch = text.charAt(offset++);
+            if (ch <= 0x7F) {
+                 if (escCodes[ch] == 0) {
+                     outputBuffer[outputPtr++] = (byte) ch;
+                     continue;
+                 }
+                 int escape = escCodes[ch];
+                 if (escape > 0) { // 2-char escape, fine
+                     outputBuffer[outputPtr++] = BYTE_BACKSLASH;
+                     outputBuffer[outputPtr++] = (byte) escape;
+                 } else {
+                     // ctrl-char, 6-byte escape...
+                     outputPtr = _writeGenericEscape(ch, outputPtr);
+                }
+                continue;
+            }
+            if (ch <= 0x7FF) { // fine, just needs 2 byte output
+                outputBuffer[outputPtr++] = (byte) (0xc0 | (ch >> 6));
+                outputBuffer[outputPtr++] = (byte) (0x80 | (ch & 0x3f));
+            } else {
+                outputPtr = _outputMultiByteChar(ch, outputPtr);
+            }
+        }
+        _outputTail = outputPtr;
+    }
+    
     /*
     /**********************************************************
     /* Internal methods, low-level writing, text segment
@@ -1307,8 +1341,7 @@ public class UTF8JsonGenerator
      * Same as <code>_writeStringSegment2(char[], ...)</code., but with
      * additional escaping for high-range code points
      */
-    private final void _writeStringSegmentASCII2(final char[] cbuf, int offset, final int end)
-        throws IOException, JsonGenerationException
+    private final void _writeStringSegmentASCII2(final char[] cbuf, int offset, final int end) throws IOException
     {
         // Ok: caller guarantees buffer can have room; but that may require flushing:
         if ((_outputTail +  6 * (end - offset)) > _outputEnd) {
@@ -1352,6 +1385,50 @@ public class UTF8JsonGenerator
         _outputTail = outputPtr;
     }
 
+    private final void _writeStringSegmentASCII2(final String text, int offset, final int end) throws IOException
+    {
+        // Ok: caller guarantees buffer can have room; but that may require flushing:
+        if ((_outputTail +  6 * (end - offset)) > _outputEnd) {
+            _flushBuffer();
+        }
+    
+        int outputPtr = _outputTail;
+    
+        final byte[] outputBuffer = _outputBuffer;
+        final int[] escCodes = _outputEscapes;
+        final int maxUnescaped = _maximumNonEscapedChar;
+        
+        while (offset < end) {
+            int ch = text.charAt(offset++);
+            if (ch <= 0x7F) {
+                 if (escCodes[ch] == 0) {
+                     outputBuffer[outputPtr++] = (byte) ch;
+                     continue;
+                 }
+                 int escape = escCodes[ch];
+                 if (escape > 0) { // 2-char escape, fine
+                     outputBuffer[outputPtr++] = BYTE_BACKSLASH;
+                     outputBuffer[outputPtr++] = (byte) escape;
+                 } else {
+                     // ctrl-char, 6-byte escape...
+                     outputPtr = _writeGenericEscape(ch, outputPtr);
+                 }
+                 continue;
+            }
+            if (ch > maxUnescaped) { // [JACKSON-102] Allow forced escaping if non-ASCII (etc) chars:
+                outputPtr = _writeGenericEscape(ch, outputPtr);
+                continue;
+            }
+            if (ch <= 0x7FF) { // fine, just needs 2 byte output
+                outputBuffer[outputPtr++] = (byte) (0xc0 | (ch >> 6));
+                outputBuffer[outputPtr++] = (byte) (0x80 | (ch & 0x3f));
+            } else {
+                outputPtr = _outputMultiByteChar(ch, outputPtr);
+            }
+        }
+        _outputTail = outputPtr;
+    }
+    
     /*
     /**********************************************************
     /* Internal methods, low-level writing, text segment
@@ -1363,8 +1440,7 @@ public class UTF8JsonGenerator
      * Same as <code>_writeStringSegmentASCII2(char[], ...)</code., but with
      * additional checking for completely custom escapes
      */
-    private final void _writeCustomStringSegment2(final char[] cbuf, int offset, final int end)
-        throws IOException, JsonGenerationException
+    private final void _writeCustomStringSegment2(final char[] cbuf, int offset, final int end) throws IOException
     {
         // Ok: caller guarantees buffer can have room; but that may require flushing:
         if ((_outputTail +  6 * (end - offset)) > _outputEnd) {
@@ -1421,6 +1497,63 @@ public class UTF8JsonGenerator
         _outputTail = outputPtr;
     }
 
+    private final void _writeCustomStringSegment2(final String text, int offset, final int end) throws IOException
+    {
+        // Ok: caller guarantees buffer can have room; but that may require flushing:
+        if ((_outputTail +  6 * (end - offset)) > _outputEnd) {
+            _flushBuffer();
+        }
+        int outputPtr = _outputTail;
+    
+        final byte[] outputBuffer = _outputBuffer;
+        final int[] escCodes = _outputEscapes;
+        // may or may not have this limit
+        final int maxUnescaped = (_maximumNonEscapedChar <= 0) ? 0xFFFF : _maximumNonEscapedChar;
+        final CharacterEscapes customEscapes = _characterEscapes; // non-null
+        
+        while (offset < end) {
+            int ch = text.charAt(offset++);
+            if (ch <= 0x7F) {
+                 if (escCodes[ch] == 0) {
+                     outputBuffer[outputPtr++] = (byte) ch;
+                     continue;
+                 }
+                 int escape = escCodes[ch];
+                 if (escape > 0) { // 2-char escape, fine
+                     outputBuffer[outputPtr++] = BYTE_BACKSLASH;
+                     outputBuffer[outputPtr++] = (byte) escape;
+                 } else if (escape == CharacterEscapes.ESCAPE_CUSTOM) {
+                     SerializableString esc = customEscapes.getEscapeSequence(ch);
+                     if (esc == null) {
+                         _reportError("Invalid custom escape definitions; custom escape not found for character code 0x"
+                                 +Integer.toHexString(ch)+", although was supposed to have one");
+                     }
+                     outputPtr = _writeCustomEscape(outputBuffer, outputPtr, esc, end-offset);
+                 } else {
+                     // ctrl-char, 6-byte escape...
+                     outputPtr = _writeGenericEscape(ch, outputPtr);
+                 }
+                 continue;
+            }
+            if (ch > maxUnescaped) { // [JACKSON-102] Allow forced escaping if non-ASCII (etc) chars:
+                outputPtr = _writeGenericEscape(ch, outputPtr);
+                continue;
+            }
+            SerializableString esc = customEscapes.getEscapeSequence(ch);
+            if (esc != null) {
+                outputPtr = _writeCustomEscape(outputBuffer, outputPtr, esc, end-offset);
+                continue;
+            }
+            if (ch <= 0x7FF) { // fine, just needs 2 byte output
+                outputBuffer[outputPtr++] = (byte) (0xc0 | (ch >> 6));
+                outputBuffer[outputPtr++] = (byte) (0x80 | (ch & 0x3f));
+            } else {
+                outputPtr = _outputMultiByteChar(ch, outputPtr);
+            }
+        }
+        _outputTail = outputPtr;
+    }
+
     private final int _writeCustomEscape(byte[] outputBuffer, int outputPtr, SerializableString esc, int remainingChars)
         throws IOException, JsonGenerationException
     {
@@ -1761,8 +1894,7 @@ public class UTF8JsonGenerator
         return inputOffset;
     }
 
-    protected final void _outputSurrogates(int surr1, int surr2)
-        throws IOException
+    protected final void _outputSurrogates(int surr1, int surr2) throws IOException
     {
         int c = _decodeSurrogate(surr1, surr2);
         if ((_outputTail + 4) > _outputEnd) {
@@ -1788,13 +1920,18 @@ public class UTF8JsonGenerator
     {
         byte[] bbuf = _outputBuffer;
         if (ch >= SURR1_FIRST && ch <= SURR2_LAST) { // yes, outside of BMP; add an escape
-            bbuf[outputPtr++] = BYTE_BACKSLASH;
-            bbuf[outputPtr++] = BYTE_u;
-            
-            bbuf[outputPtr++] = HEX_CHARS[(ch >> 12) & 0xF];
-            bbuf[outputPtr++] = HEX_CHARS[(ch >> 8) & 0xF];
-            bbuf[outputPtr++] = HEX_CHARS[(ch >> 4) & 0xF];
-            bbuf[outputPtr++] = HEX_CHARS[ch & 0xF];
+            // 23-Nov-2015, tatu: As per [core#223], may or may not want escapes;
+            //   it would be added here... but as things are, we do not have proper
+            //   access yet...
+//            if (Feature.ESCAPE_UTF8_SURROGATES.enabledIn(_features)) {
+                bbuf[outputPtr++] = BYTE_BACKSLASH;
+                bbuf[outputPtr++] = BYTE_u;
+                
+                bbuf[outputPtr++] = HEX_CHARS[(ch >> 12) & 0xF];
+                bbuf[outputPtr++] = HEX_CHARS[(ch >> 8) & 0xF];
+                bbuf[outputPtr++] = HEX_CHARS[(ch >> 4) & 0xF];
+                bbuf[outputPtr++] = HEX_CHARS[ch & 0xF];
+//            } else { ... }
         } else {
             bbuf[outputPtr++] = (byte) (0xe0 | (ch >> 12));
             bbuf[outputPtr++] = (byte) (0x80 | ((ch >> 6) & 0x3f));
@@ -1802,7 +1939,7 @@ public class UTF8JsonGenerator
         }
         return outputPtr;
     }
-    
+
     private final void _writeNull() throws IOException
     {
         if ((_outputTail + 4) >= _outputEnd) {
@@ -1817,8 +1954,7 @@ public class UTF8JsonGenerator
      * 
      * @param charToEscape Character to escape using escape sequence (\\uXXXX)
      */
-    private int _writeGenericEscape(int charToEscape, int outputPtr)
-        throws IOException
+    private int _writeGenericEscape(int charToEscape, int outputPtr) throws IOException
     {
         final byte[] bbuf = _outputBuffer;
         bbuf[outputPtr++] = BYTE_BACKSLASH;
